# -*- coding: utf-8 -*-
"""
Created on Fri Mar  8 13:51:45 2019

@author: HugoFara

Module proposing different function optimization algorithms. The output is
generally leg dimensions of walking linkages.
"""
from copy import deepcopy
import numpy as np
from numpy.random import rand
from numpy.linalg import norm
# Particle swarm optimization
from pyswarms.single.local_best import LocalBestPSO
from pyswarms.utils.plotters import plot_cost_history

from .geometry import sqr_dist


def variator(ite, delta_dim, min_dim=5, max_dim=5):
    """
    Return an iterable of all possibles variations of elements.

    Number of variations: ((max_dim - 1 / min_dim) / delta_dim) ** len(ite).

    Because linkage are not tolerant to violent changes, the order of output
    for the coefficients is very important.

    The coefficient is in order: middle → min (step 2), min → middle (step 2),
    middle → max (step 1), so that there is no huge variation.

    Parameters
    ----------
    ite : sequence of floats
        Elements that should vary.
    delta_dim : float
        Scale factor for each variation.
    min_dim : float, optional
        Minimal scale reduction (dimensions not shorter that dim/min_dim).
        The default is 5.
    max_dim : float, optional
        maximal scale augmentation (not above dim * max_dim). The default is 5.

    Returns
    -------
    Generator
        Each element is the list of floats with little variations.

    """
    # We copy the sequence
    c = tuple(ite)
    inv = 1 / min_dim
    variations = int((max_dim - inv) / delta_dim)
    coef = tuple(inv + i * delta_dim for i in range(variations))
    middle, even = variations // 2, variations % 2
    # We reordinate the coo
    coef = (coef[middle:0:-2] + coef[0:1]
            + coef[even:middle:2] + coef[middle+1:])
    return recurs_variator(c, list(c), coef)


def recurs_variator(ite, copy, coef_list, num=0):
    """
    Recursive dimensions generator.

    Called by variator. Only "copy" is modified.
    """
    for coef in coef_list:
        copy[num] = ite[num] * coef
        yield copy
        if num < len(ite) - 1:
            for j in recurs_variator(ite, copy, coef_list, num + 1):
                yield j


def exhaustive_optimization(eval_func, linkage, parameters, n_results=10,
                            delta_dim=.5, min_dim=2, max_dim=2):
    """
    Return the list of dimensions optimizing eval_func.

    We start wy making the dimensions vary, then we try a crank revolution on
    10 points. If no error we try on 75 points (higher precision).

    Each dimensions set has a score, which is added in an array of n_results
    results, containg the linkages with best scores.

    Arguments
    ---------
    * eval_func: evaluation function. Function of R^len(linkage.joints) → R.
    * linkage: linkage to evaluate
    * parameters: parameters that will be modified. Geometric constraints.
    * n_results: number of best cancidates to keep

    """
    n = (max_dim - 1 / min_dim) / delta_dim
    print("Computation running, about {} combinations...".format(
        int(n) ** len(parameters)))
    # Results to output: scores, dimensions and initial positions
    # scores will be in decreasing order
    results = [[-float('inf'), [], []] for i in range(n_results)]
    prev = [i.coord() for i in linkage.joints]
    # We start by a "fall" : we do not want to break the system by modifying
    # dimensions, so we assess it is normally behaving, and we change
    # dimensions progressively to minimal dimensions.
    # A list of dictionaries of all possible dimensions
    for dim in variator(parameters, delta_dim, min_dim, max_dim):
        # Check performances
        score = eval_func(linkage, dim, prev)
        for r in results:
            if r[0] < score:
                r[0] = score
                r[1] = dim.copy()
                r[2] = prev.copy()
                break
        if score and False:
            # Save initial positions if score not null, for further
            # computations
            prev = [k.coord() for k in linkage.joints]
    return results


def particle_swarm_optimization(eval_func, linkage, begin, n_indi=21,
                                delta_dim=.3, iner=.8, leader=.6, follower=.8,
                                neigh=1, lifetime=0, blind_iter=5, merge=.1,
                                ite=1000, iterable=False, **kwargs):
    """
    Particle Swarm Optimization wrapper for pyswarms.

    This function is a simple wrapper to optimize a linkage using PSO. It will
    mainly call the LocalBestPSO function from pyswarms.single.

    Parameters
    ----------
    eval_func : callable, must return float
        The evaluation function.
        Input: a set of dimensions and intial position for the linkage
        Output: a score, a float
        The swarm will look for the HIGHEST score.
    linkage : Linkage
        Linkage to be optimized. Make sure to give an optimized linkage for
        better results
    begin : tuple of float
        initial dimensions.
    n_indi : float, optional
        Number of particles in the swarm. The default is 21.
    delta_dim : float, optional
        Unused, for legacy purposes. The default is .3.
    iner : float, optional
        Inertia of each particle, w in pyswarms. The default is .8.
    leader : float, optional
        Learning coefficient of each particle, c1 in pyswarms.
        The default is .6.
    follower : float, optional
        Social coefficient, c2 in pyswarms. The default is .8.
    neigh : float, optional
        With the Euclidian norm, distance under a particle is a neighbour.
        The default is 1.
    lifetime : TYPE, optional
        DESCRIPTION. The default is 0.
    blind_iter : TYPE, optional
        DESCRIPTION. The default is 5.
    merge : TYPE, optional
        DESCRIPTION. The default is .1.
    ite : TYPE, optional
        DESCRIPTION. The default is 1000.
    iterable : TYPE, optional
        DESCRIPTION. The default is False.
    **kwargs : TYPE
        DESCRIPTION.

    Returns
    -------
    list
        List of 3 elements: best dimensions, best score and initial positions.

    """
    dims_len = len(begin)
    # A distribution not too far from initial position
    init_pos = np.random.normal(loc=1., scale=.05,
                                size=(n_indi, dims_len)) * begin
    options = {'c1': leader, 'c2': follower, 'w': iner, 'p': 1, 'k': 5}
    pos = tuple(j.coord() for j in linkage.joints)
    optimizer = LocalBestPSO(
        n_particles=n_indi, dimensions=dims_len, options=options,
        init_pos=init_pos, **kwargs)
    # vectorized_eval_func=np.vectorize(eval_func, signature='(n),(m,k)->()')               # First optimizer with blind iterations
    optimizer.optimize(
        #vectorized_eval_func,
        lambda dims, pos: np.any(dims < np.zeros(len(dims))) * float('inf'),
        blind_iter, pos=pos)
    optimizer.reset()
    out = optimizer.optimize(
        #vectorized_eval_func,
        lambda dims, pos: np.array([-eval_func(d, pos) for d in dims]),
        ite, pos=pos)
    plot_cost_history(optimizer.cost_history)
    return [(out[0], out[1], pos)]


class Particle(object):
    """Simple particle swarm optimization (PSO) class."""

    __slots__ = ("pos", "speed", "eval", "ini", "swarm",
                 "index", "max_dist", "score", "best_score",
                 "best_pos", "dists", "neighbours")

    def __init__(self, pos, speed, ini, eva, swarm, max_dist=float('inf')):
        """
        Arguments:
        pos: particule position
        speed: intial velocity
        ini: additionnal data (initial position of the linkage)
        eva: evaluation function
        swarm: swarm the particule belongs to
        max_dist: maximum distance at which another particule can attract this one
        """
        self.pos = np.array(pos)
        self.speed = np.array(speed)
        self.ini, self.eval = ini, eva
        self.swarm = swarm
        self.index = len(self.swarm)
        self.swarm.append(self)
        self.max_dist = max_dist
        self.best_score = -float('inf')
        self.score = -float('inf')
        self.new_score()
        self.best_pos, self.best_score = self.pos, self.score
        self.dists = np.empty(len(self.swarm) * (len(self.swarm) - 1) // 2)
        self.neighbours = tuple()

    def __get_row__(self):
        """Get the row in swarm index."""
        return (self.index * (len(self.swarm) - 3)
                - (self.index - 1) * (self.index - 2) // 2)

    def __del__(self):
        """Update swarm indexes."""
        if len(self.swarm) > self.index:
            self.swarm.pop(self.index)
            for p in self.swarm[self.index:]:
                p.index -= 1
        self.update_dists()
        for p in self.swarm:
            p.update_neighbours()

    """ Methodes affecting single agent"""
    def new_score(self):
        """ Gets the score at the current position and updates best_scores and
        best_position if the new score is superior OR equal to the previous
        one """
        self.score = self.eval(self.pos, self.ini)
        # Large equality important when score is null.
        # (avoid staying on null position)
        if self.score >= self.best_score:
            self.best_pos, self.best_score = self.pos, self.score
        return self.score

    def update_neighbours(self):
        """Update this particle list of neighbours."""
        row = self.__get_row__()
        self.neighbours = tuple(
                filter(
                        lambda p: (p is not self
                                   and self.dists[row + p.index] < self.max_dist ** 2),
                                   self.swarm))

    def move(self, blind=False):
        """
        Move the particle.

        Start by updating position using speed, then update particles data.
        """
        self.pos += self.speed
        if blind:
            return
        self.update_neighbours()
        self.score = self.eval(self.pos, self.ini)
        # Large equality important when score is null.
        # (avoid staying on null position)
        if self.score >= self.best_score:
            self.best_pos, self.best_score = self.pos, self.score
        # update_dists()
        # new_score()

    def die(self):
        """Destroy particle. Equivalent to __del__."""
        self.__del__()

    # Methods affecting whole swarm.

    def update_dists(self):
        """Update the list of distances relative to this particle."""
        l_s = len(self.swarm)
        dists = np.empty(l_s * (l_s - 1) // 2)
        # TODO: IN PARALLEL
        for p1 in self.swarm:
            p1.dists = dists
            row = p1.__get_row__()
            for p2 in self.swarm[self.index + 1:]:
                if row + p2.index >= len(dists):
                    raise Warning(
                            "Index overflow!\nrow: {}"
                            ", index: {}, len(dists):  {}".format(
                                    row, p2.index, len(dists)))
                dists[row + p2.index] = sqr_dist(p1.pos, p2.pos)
                p2.dists = self.dists

    def kill_swarm(self):
        print("Decimation")
        #del dists
        for p in self.swarm:
            del p.dists
        self.swarm.clear()


def particle_swarm_optimization_legacy(
        eval_func, linkage, begin, n_indi=21, delta_dim=.3, iner=.8, leader=.6,
        follower=.8, neigh=1, lifetime=0, blind_iter=5, merge=.1,
        ite=1000, iterable=False):
    """
    Particle swarm optimization. We look for best solutions to eval_func,
    starting from a begin set with real values. We use a particle swarm of
    n_indi agents.

    Return an iterable of list of agents' position.

    Parameters:
        - eval_func: function giving the score to each position
        - linkage: linka["linkage"]
        - begin: iterable of initial positions (dimensions of the linkage)
        - n_indi: nomber of agents in swarm
        - min_dim: each dimension is divised by min_dim. If an agent is below
        begin[i] / min_dim, his score is set to 0
        - max_dim: upper bound, agents above begin[i] * max_dim have score 0
        - iner: agents inertia
        - leader: trust is self best position, learning coefficient
        - follower: social coefficient
        - lifetime: if an agent has score 0 for lifetime frames, he dies.
    """
    c = [i.coord() for i in linkage.joints]
    swarm = []
    Particle(begin.copy(), 0, c.copy(), eval_func, swarm, neigh)
    for p in range(n_indi - 1):
        # They run away for initial position at given speed
        v = rand(len(begin)) - .5
        Particle(begin.copy(), delta_dim * v, deepcopy(c),
                 eval_func, swarm, neigh)
    swarm[0].update_dists()

    #print("Identity : ", swarm[0] is swarm[-1])
    if iterable:
        yield [(p.pos, p.score, p.ini) for p in swarm], -1
    c = [0] * len(swarm)
    #print("agents : ", len(swarm))
    # We start calulations
    for j in range(ite):
        # Save intial positions
        c[:] = [p.pos.copy() for p in swarm]
        # Update parameters
        for p in swarm:
            # Update agent position
            p.move(j < blind_iter)
            if j < blind_iter:
                v = rand(len(begin)) - .5
                p.speed = delta_dim / norm(v) * v
                #if p.score:
                #    p.ini = [i.coord() for i in linkage]
                continue
            # Best position in neighborhood
            pg = np.array(max(p.neighbours, key=lambda x: x.score, default=p).pos)
            # print("Does" + ("" if max(filter(dists[i]) == pg else "not") + " work"))
            # Update speed
            # USE norm(pg - p.pos), norm(p.best_bos - p.pos), check stability
            p.speed = (iner * p.speed
                       + leader * rand() * (pg - p.pos)
                       + follower * rand() * (p.best_pos - p.pos))
            # Score evaluation
            if p.new_score():
                p.ini = [i.coord() for i in linkage.joints]
        if merge and j > blind_iter:
            for p in swarm:
                # TODO: add a second axis to self.dists
                deletion = False
                for k in filter(
                        lambda x: sqr_dist(p.speed, x.speed) < merge ** 2,
                        p.neighbours):
                    print('Merge mark')
                    # Least score particule to be deleted
                    (p if p.score < k.score else k).__del__()
                    #del swarm[max(p.index, b_score_part.index)]
                    #swarm[min(p.index, b_score_part.index)] = b_score_part
                    deletion = True
                    break
                if deletion:
                    #p.update_dists()
                    #p.update_neighbours()
                    break

        if iterable:
            # yield map(lambda p: (p.pos, p.score, p.ini), swarm), j
            yield [(p.pos, p.score, p.ini) for p in swarm], j
    if not iterable:
        coco = [(p.pos, p.score, p.ini) for p in swarm]
    # swarm detruction
    #swarm[0].kill_swarm()
    if not iterable:
        return coco



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed May  5 17:34:45 2021

@author: HugoFara

General geometry functions.
"""
import math


def dist_builtin(point1, point2):
    """
    Euclidian distance between two 2D points.

    Legacy built-in unoptimized equivalent of math.dist in Python 3.8.
    """
    return math.sqrt((point1[0] - point2[0]) ** 2
                     + (point1[1] - point2[1]) ** 2)


if hasattr(math, 'dist'):
    dist = math.dist
else:
    print('Unable to import dist from math. Using built-in function.')
    dist = dist_builtin


def sqr_dist(point1, point2):
    """
    Square of the distance between two points.

    Faster than dist.
    """
    return (point1[0] - point2[0]) ** 2 + (point1[1] - point2[1]) ** 2


def norm(vec):
    """Return the norm of a 2-dimensional vector."""
    return math.sqrt(vec[0] ** 2 + vec[1] ** 2)


def cyl_to_cart(radius, theta, ori=(0, 0)):
    """
    Convert polar coordinates into cartesian.

    Arguments
    ---------
    radius: distance from ori
    theta: angle is the angle starting from abscisses axis
    ori: origin point.
    """
    return radius * math.cos(theta) + ori[0], radius * math.sin(theta) + ori[1]


def __secant_circles_intersections__(distance, dist_x, dist_y, mid_dist,
                                     radius1, projected):
    """Return the TWO intersections of secante circles."""
    # Distance between projected P and points
    # and the points of which P is projection
    height = math.sqrt(radius1 ** 2 - mid_dist ** 2) / distance
    inter1 = (
        projected[0] + height * dist_y,
        projected[1] - height * dist_x
    )
    inter2 = (
        projected[0] - height * dist_y,
        projected[1] + height * dist_x
    )
    return 2, inter1, inter2


def circle_intersect(circle1, circle2, tol=0.0):
    """
    Return the intersections between two circles.

    Transcription of a Matt Woodhead program, method provided by Paul Bourke,
    1997. http://paulbourke.net/geometry/circlesphere/.

    Arguments
    ---------
    circle1: first circle, sequence of (abscisse, ordinate, radius)
    circle2: second circle, sequence of (abscisse, ordinate, radius)
    tol: distance under which two points are considered equal.
    """
    x_1, y_1, radius1 = circle1
    x_2, y_2, radius2 = circle2

    dist_x, dist_y = x_2 - x_1, y_2 - y_1
    # Distance between circles centers
    distance = math.sqrt(dist_x ** 2 + dist_y ** 2)
    if distance > radius1 + radius2:
        # Circles two far
        return (0, )
    if distance < abs(radius2 - radius1):
        # One circle in the other
        return (0, )
    if distance <= tol and abs(radius1 - radius2) <= tol:
        # Same circle
        return 3, circle1

    dual = True
    if abs(abs(radius1 - distance) - radius2) <= tol:
        # Tangent circles
        dual = False

    # Distance from first circle's center to orthogonal projection
    # of circles intersections, on the axis between circles' centers
    mid_dist = (radius1 ** 2 - radius2 ** 2 + distance ** 2) / (2 * distance)
    # projected is easy now
    projected = (
        x_1 + (mid_dist * dist_x) / distance,
        y_1 + (mid_dist * dist_y) / distance
    )

    if dual:
        return __secant_circles_intersections__(distance, dist_x, dist_y,
                                                mid_dist, radius1, projected)
    return 1, projected


def intersection(obj_1, obj_2, tol=0.0):
    """
    Return intersection between two objects (points or circles).

    Return nothing of no intersection.
    tol: absolute tolerance to use if provided.
    """
    # Two points
    if len(obj_1) == 2 and len(obj_2) == 2:
        if obj_1 == obj_2 or tol and dist(obj_1, obj_2) <= tol:
            return obj_1
        return
    # Two circles
    if len(obj_1) == 3 and len(obj_2) == 3:
        return circle_intersect(obj_1, obj_2)[1:]
    # Point and circle
    if len(obj_1) == 2 and len(obj_2) == 3:
        if dist(obj_1, obj_2[:2]) - obj_2[2] <= tol:
            return obj_1
        return
    # Circle and point
    if len(obj_1) == 3 and len(obj_2) == 2:
        return intersection(obj_2, obj_1, tol)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Apr 16 16:39:21 2021

@author: HugoFara
"""
from math import atan2, gcd, tau
from abc import ABC
from .geometry import sqr_dist, circle_intersect, cyl_to_cart


class UnbuildableError(Exception):
    """Should be raised when the constraints cannont be solved."""

    def __init__(self, joint, message='Unable to solve constraints'):
        self.joint = joint
        super().__init__(message)

    def __str__(self):
        return f"{self.joint} cannot be solved"


class HypostaticError(Exception):
    """Exception when the system can be solved in multiple ways."""

    def __init__(self, linkage, message='The system is hypostatic!'):
        self.linkage = linkage
        super().__init__(message)


class Joint(ABC):
    """
    Geometric constraint expressed by two joints.

    Abstract class, should not be used by itself.
    """

    __slots__ = "x", "y", "joint0", "joint1", "name"

    def __init__(self, x=0, y=0, joint0=None, joint1=None, name=None):
        """
        Create Joint.

        Arguments
        ---------
        * x: joint axis, scalar
        * y: joint ordinate
        * joint0: first reference to use, usually a Joint
        * joint1: second reference, same behavior as joint0
        * name: unique name
        """
        self.x, self.y = x, y
        self.joint0, self.joint1 = joint0, joint1
        self.name = name or str(id(self))

    def __repr__(self):
        """Represent object with class name, coordinates, name and state."""
        return "{}(x={}, y={}, name={})".format(self.__class__.__name__,
                                                self.x, self.y, self.name)

    def __get_joints__(self):
        """Return constraint joints as a tuple."""
        return self.joint0, self.joint1

    def coord(self):
        """Return cartesian coordinates."""
        return self.x, self.y

    def set_coord(self, *args):
        """Take a sequence or two scalars, and assign them to object x, y."""
        if len(args) == 1:
            self.x, self.y = args[0]
        else:
            self.x, self.y = args[0], args[1]


class Static(Joint):
    """
    Special case of Joint that should not move.

    Mostly used for the frame.
    """

    __slots__ = tuple()

    def __init__(self, x=0, y=0, name=None):
        super().__init__(x, y, name=name)

    def reload(self):
        """Do nothing, for consistence only."""
        return

    def set_constraints(self, *args):
        """Do nothing, for consistence only."""
        return

    def set_anchor0(self, joint):
        self.joint0 = joint

    def set_anchor1(self, joint):
        self.joint1 = joint


class Fixed(Joint):
    """Define a joint using parents locations only, with no ambiguity."""

    __slots__ = "r", "angle"

    def __init__(self, x=None, y=None, joint0=None, joint1=None,
                 distance=None, angle=None, name=None):
        """
        Create a point, of position fully defined by its two references.

        Arguments
        ---------
        * joint0: first reference joint,
        * joint1: second reference joint.
        * distance: to keep constant between joint0 and self.
        * angle: It is the angle (joint1, joint0, self).
        Should be in radian and in trigonometric order.
        """
        super().__init__(x, y, joint0, joint1, name)
        self.angle = angle
        self.r = distance

    def reload(self):
        """
        Compute point coordinates.

        We know point position relative to its two parents, which gives a local
        space. The know the orientation of local space, so we can solve the
        whole. Local space is defined by link[0] as the origin and
        (link[0], link[1]) as abscisses axis.
        """
        if self.joint0 is None:
            return
        if self.joint0 is None or self.joint1 is None:
            raise HypostaticError(f'Not enough constraints for {self}')
        # Rotation angle of local space relative to global
        rot = atan2(self.joint1.y - self.joint0.y,
                    self.joint1.x - self.joint0.x)
        # Position in global space
        self.x, self.y = cyl_to_cart(self.r, self.angle + rot,
                                     self.joint0.coord())

    def set_constraints(self, distance=None, angle=None):
        """Set geometric constraints."""
        self.r, self.angle = distance or self.r, angle or self.angle

    def set_anchor0(self, joint, distance=None, angle=None):
        self.joint0 = joint
        self.set_constraints(distance, angle)

    def set_anchor1(self, joint):
        self.joint1 = joint


class Pivot(Joint):
    """Center of pivot joint."""

    __slots__ = "r0", "r1"

    def __init__(self, x=0, y=0, joint0=None, joint1=None, distance0=None,
                 distance1=None, name=None):
        """
        Set point position, parents, and if it is fixed for this turn.

        Arguments:
        ---------
        * x: position on horizontal axis
        * y: position on vertical axis
        * name: friendly name for human readability
        * joint0: linked pivot joint 1 (geometric constraints). A Joint
        representing the center of a pivot joint.
        * joint1: other pivot joint linked.
        * distance0: distance from joint0 to the current Joint
        * distance1: distance from joint1 to the current Joint.
        """
        super().__init__(x, y, joint0, joint1, name)
        self.r0, self.r1 = distance0, distance1

    def __get_joint_as_circle__(self, index):
        """Return reference as (ref.x, ref.y, ref.distance)."""
        if index == 0:
            return self.joint0.x, self.joint0.y, self.r0
        if index == 1:
            return self.joint1.x, self.joint1.y, self.r1
        raise ValueError(f'index should be 0 or 1, not {index}')

    def circle(self, joint):
        """
        Return first link between self and parent as a circle.

        Circle is a tuple (abscisse, ordinate, radius).
        """
        if self.joint0 is joint:
            return joint.x, joint.y, self.ref1[1]
        elif self.joint1 is joint:
            return joint.x, joint.y, self.ref2[1]
        raise ValueError(f'{joint} is not in joints of {self}')

    def reload(self):
        """Compute position of pivot joint, use the two linked joints."""
        if self.joint0 is None:
            return
        # Fixed joint as reference. In links, we only keep fixed objects
        ref = tuple(x for x in self.__get_joints__() if None not in x.coord())
        if len(ref) == 0:
            # Don't modify coordinates (irrelevant)
            return
        if len(ref) == 1:
            raise Warning("Unable to set coordinates of pivot joint {}:"
                          "Only one constraint is set."
                          "Coordinates unchanged".format(self.name))
        elif len(ref) == 2:
            # Most common case, optimized here
            coco = circle_intersect(self.__get_joint_as_circle__(0),
                                    self.__get_joint_as_circle__(1))
            if coco[0] == 0:
                raise UnbuildableError(self)
            if coco[0] == 1:
                self.x, self.y = coco[1]
            elif coco[0] == 2:
                if sqr_dist(self.coord(), coco[1]
                            ) < sqr_dist(self.coord(), coco[2]):
                    self.x, self.y = coco[1]
                else:
                    self.x, self.y = coco[2]
            elif coco[0] == 3:
                raise Warning(f"Joint {self.name} has an infinite number of"
                              "solutions, position will be arbitrary")
                # We project position on circle of possible positions
                vect = ((j-i)/abs(j-i) for i, j in zip(coco[1],
                                                       self.coord()))
                self.x, self.y = [i + j * coco[1][2] for i, j in zip(coco[1],
                                  vect)]

    def set_constraints(self, distance0=None, distance1=None):
        """Set geometric constraints."""
        self.r0, self.r1 = distance0 or self.r0, distance1 or self.r1

    def set_anchor0(self, joint, distance=None):
        self.joint0 = joint
        self.set_constraints(distance0=distance)

    def set_anchor1(self, joint, distance=None):
        self.set_constraints(distance1=distance)


class Crank(Joint):
    """Define a crank joint."""

    __slots__ = "r", "angle"

    def __init__(self, x=None, y=None, joint0=None,
                 distance=None, angle=None, name=None):
        """
        Define a crank (circular motor).

        Parameters
        ----------
        x : float, optional
            initial horizontal position, won't be used thereafter.
            The default is None.
        y : float, optional
            initial vertical position. The default is None.
        joint0 : Joint, optional
            first reference joint. The default is None.
        distance : float, optional
            distance to keep between joint0 and self. The default is None.
        angle : float, optional
            It is the angle (horizontal axis, joint0, self).
            Should be in radian and in trigonometric order.
            The default is None.
        name : str, optional
            user-friendly name. The default is None.

        Returns
        -------
        None.

        """
        super().__init__(x, y, joint0, name=name)
        self.r, self.angle = distance, angle

    def reload(self, dt=1):
        """Make a step of crank."""
        if self.joint0 is None:
            return
        if None in self.joint0.coord():
            raise HypostaticError(f'{self.joint0} has None coordinates. '
                                  f'{self} cannot be calculated')
        # Rotation angle of local space relative to global
        rot = atan2(self.y - self.joint0.y, self.x - self.joint0.x)
        self.x, self.y = cyl_to_cart(self.r, rot + self.angle * dt,
                                     self.joint0.coord())

    def set_constraints(self, distance=None, *args):
        """Set geometric constraints, only self.r is affected."""
        self.r = distance or self.r

    def set_anchor0(self, joint, distance=None):
        self.joint0 = joint
        self.set_constraints(distance=distance)


class Linkage():
    """
    A linkage is a set of Joint objects.

    It is defined kinematicaly. Coordinates are given relative to its own
    base.
    """

    __slots__ = "name", "joints", "_cranks", "_solve_order"

    def __init__(self, joints, order=None, name=None):
        """
        Define a linkage, a set of joints.

        Arguments
        ---------
        * joints: all Joint to be part of the linkage
        * order: sequence that manually define resolution order for each step.
        Should be elements of joints if provided.
        * name: linkage name
        """
        self.name = str(name or id(self))
        self.joints = tuple(joints)
        self._cranks = tuple(j for j in joints if isinstance(j, Crank))
        if order:
            self._solve_order = tuple(order)

    def __set_solve_order__(self, order):
        """Set constraints resolution order."""
        self._solve_order = order

    def __find_solving_order__(self):
        """Automatically finds solving order."""
        # TODO : test it
        solvable = [j for j in self.joints if isinstance(j, Static)]
        # True of new joints where added in the current pass
        solved_in_pass = True
        while len(solvable) < len(self.joints) and solved_in_pass:
            solved_in_pass = False
            for j in self.joints:
                if isinstance(j, Static) or j in solvable:
                    continue
                if j.joint0 in solvable:
                    if isinstance(j, Crank):
                        solvable.append(j)
                        solved_in_pass = True
                    elif j.joint1 in solvable:
                        solvable.append(j)
                        solved_in_pass = True
        if len(solvable) < len(self.joints):
            raise HypostaticError(
                'Unable to determine automatic order!'
                'Those joints are left unsolved:'
                ','.join(j for j in self.joints if j not in solvable)
                )
        self._solve_order = tuple(solvable)
        raise NotImplementedError('Unable to determine automatic order')
        return self._solve_order

    def rebuild(self, pos=None):
        """
        Redifine linkage joints and given intial positions to joints.

        pos: a tuple of initial positions, in the same order as linka["order"],
        for linka["crank"] at position (0, -1). Fixed_Joint WON'T be modified.
        Coordinates does not need to be precise, they will allow us the best
        fitting position between all possible positions satifying constraints.
        """
        if not hasattr(self, '_solve_order'):
            self.__find_solving_order__()

        # Links parenting in descending order solely.
        # Parents joint do not have children.
        if pos is not None:
            # Defintition of initial coordinates
            for p, j in zip(pos, self.joints):
                j.set_coord(p)

    def get_pos(self):
        """Return positions of all elements of the system."""
        return (j.coord() for j in self.joints)

    def hyperstaticity(self):
        """Return the hyperstaticity degree of the linkage in 2D."""
        raise NotImplementedError('Not added in this version!')

    def step(self, iterations=None, dt=1):
        """

        Make a step of the linkage.

        Parameters
        ----------
        iterations : int, optional
            Number of iterations to run across.
            If None, the default is self.get_rotation_period().
        dt : int, optional
            Amount of rotation to turn the cranks by.
            All cranks rotate by their self.angle * dt. The default is 1.

        Yields
        ------
        generator
            Iterable of the joints' coordinates.

        """
        if iterations is None:
            iterations = self.get_rotation_period()
        for i in range(iterations):
            for j in self._solve_order:
                if isinstance(j, Crank):
                    j.reload(dt)
                else:
                    j.reload()
            yield tuple(j.coord() for j in self.joints)

    def set_coords(self, coords):
        """Set coordinatess for all joints of the linkage."""
        for j, c in zip(self.joints, coords):
            j.set_coord(c)

    def set_num_constraints(self, constraints):
        """
        Set numeric constraints for this linkage.

        Numeric constraints are distances or angles between joints.

        Arguments
        ---------
        * constraints: a sequence of tuples of digits. Should be in same order
        as self.joints. Each element will be passed to the set_constraints
        method of each correspondig Joint.
        """
        for j, c in zip(self.joints, constraints):
            j.set_constraints(*c)

    def get_rotation_period(self):
        """
        Return the number of iterations to finish in the previous state.

        Formally it is the common denominator of all crank periods.

        Returns
        -------
        Number of iterations, with dt=1.

        """
        periods = 1
        for j in self.joints:
            if isinstance(j, Crank):
                freq = round(tau / j.angle)
                periods = periods * freq // gcd(periods, freq)
        return periods



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jun 14 12:13:58 2021

@author: HugoFara

This module makes visualisation of linkages easy using matplotlib.
"""
import matplotlib.pyplot as plt
import matplotlib.animation as anim

from .linkage import Crank, Fixed, Static, Pivot

# List of animations
ani = []


def plot_static_linkage(linkage, ax, locii, locus_highlights=None,
                        show_legend=False):
    """
    Plot a linkage without movement.

    Parameters
    ----------
    linkage : Linkage
        The linkage you want to see.
    ax : Artist
        The graph we should draw on.
    locii : sequence
        List of list of coordinates. They will be plotted.
    locus_highlights : list, optional
        If a list, shoud be a list of list of coordinates you want to see
        highlighted. The default is None.
    show_legend : bool, optional
        To add an automatic legend to the graph. The default is False.

    Returns
    -------
    None.

    """
    ax.set_aspect('equal')
    ax.grid(True)
    for i in range(len(linkage.joints)):
        ax.plot(tuple(j[i][0] for j in locii), tuple(j[i][1] for j in locii))
    if locus_highlights:
        for locus in locus_highlights:
            ax.scatter(tuple(coord[0] for coord in locus),
                       tuple(coord[1] for coord in i))
    if show_legend:
        ax.set_title("Individual joint locus")
        ax.set_xlabel("Points abscisses")
        ax.set_ylabel("Ordinates")
        ax.legend(tuple(i.name for i in linkage.joints[:11]))
        ax.set_xlim(min(min(i[0] for i in m) for m in locii) - 4)


def update_animated_plot(linkage, index, im, locii):
    """
    Modify im to make the animation run faster.

    Parameters
    ----------
    linkage : TYPE
        DESCRIPTION.
    index : int
        Frame index.
    im : list of images Artists
        Artist to be modified.
    locii : list
        list of locuses.

    Returns
    -------
    im : list of images Artists
        Updated version.

    """
    a = 0
    locus = locii[index]
    for j, pos in enumerate(locus):
        joint = linkage.joints[j]
        # Draw link to first parent if it exists
        if joint.joint0 is None:
            continue
        par_locus = locus[linkage.joints.index(joint.joint0)]
        im[a].set_data([par_locus[0], pos[0]], [par_locus[1], pos[1]])
        a += 1
        # Then second parent
        if isinstance(joint, (Crank, Static)):
            continue
        par_locus = locus[linkage.joints.index(joint.joint1)]
        im[a].set_data([par_locus[0], pos[0]], [par_locus[1], pos[1]])
        a += 1
    return im


def plot_animated_linkage(linkage, fig, ax, locii, frames=None, interval=.04):
    """
    Plot a linkage with an animation.

    Parameters
    ----------
    linkage : Linkage
        DESCRIPTION.
    fig : matplotlib.figure.Figure
        Figure to support the axes.
    ax : matplotlib.axes._subplots.AxesSubplot
        The subplot to draw on.
    locii : list
        list of list of coordinates.
    frames : int, optional
        Number of frames to draw the linkage on. The default is None.
    interval : float, optional
        Minimal amount of time between two frames. The default is .04 (24 fps).

    Returns
    -------
    None.

    """
    ax.set_aspect('equal')
    ax.set_title("Animation")

    im = []
    for j in linkage.joints:
        if isinstance(j, Static):
            # We will draw a fictive line with closest neighbor
            if j.joint0 is not None:
                im.append(ax.plot([], [], c='k', animated=False)[0])
        elif isinstance(j, Crank):
            # Crank has one parent only
            im.append(ax.plot([], [], c='g', animated=True)[0])
        elif isinstance(j, Fixed):
            im.append(ax.plot([], [], c='r', animated=True)[0])
            im.append(ax.plot([], [], c='r', animated=True)[0])
        elif isinstance(j, Pivot):
            im.append(ax.plot([], [], c='b', animated=True)[0])
            im.append(ax.plot([], [], c='b', animated=True)[0])

    padding = .5
    ax.set_xlim(min((min((i[0] for i in m)) for m in locii)) - padding,
                max((max((i[0] for i in m)) for m in locii)) + padding)
    ax.set_ylim(min((min((i[1] for i in m)) for m in locii)),
                max((max((i[1] for i in m)) for m in locii)) + padding)
    ani.append(anim.FuncAnimation(
        fig=fig,
        func=lambda index: update_animated_plot(linkage, index, im, locii),
        frames=frames, blit=True, interval=interval, repeat=True,
        save_count=frames))


def show_results(linkage, save=False, prev=None, L=None, title=str(len(ani)),
                 duration=5, points=100, fps=24):
    """
    Display results as an animated drawing.

    Parameters
    ----------
    linkage : pylinkage.linkage.Linkage
        The Linkage you want to draw.
    save : bool, optional
        To save the animation. The default is False.
    prev : list, optional
        Previous coordinates to use for linkage. The default is None.
    L : list, optional
        list of locii. The default is None.
    title : str, optional
        Figure title. The default is str(len(ani)).
    duration : float, optional
        Animation duration (in seconds). The default is 5.
    points : int, optional
        Number of point to draw for a crank revolution. The default is 100.
    fps : float, optional
        Number of frame per second for the output video. The default is 24.

    Returns
    -------
    None.

    """
    # Define intial positions
    linkage.rebuild(prev)
    if L is None:
        factor = int(12 / points) + 1
        L = tuple(tuple(i) for i in linkage.step(
            iterations=points * factor, dt=1 / factor))

    fig = plt.figure("Result " + title, figsize=(14, 7))
    fig.clear()
    ax1 = fig.add_subplot(1, 2, 1)
    plot_static_linkage(linkage, ax1, L, show_legend=True)
    ax2 = fig.add_subplot(1, 2, 2)
    plot_animated_linkage(linkage, fig, ax2, L)
    plt.tight_layout()
    plt.show(block=False)
    plt.pause(duration)
    plt.close()
    if save:
        writer = anim.FFMpegWriter(fps=fps, bitrate=3600)
        ani[-1].save("Linkage animated.mp4", writer=writer)
